/*
 * Time zone resources
 *
 * Copyright 2018 Akihiro Sagawa
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include "winresrc.h"

#pragma makedep po

LANGUAGE LANG_ENGLISH, SUBLANG_DEFAULT

STRINGTABLE
{
  39616 "Afghanistan Standard Time"
  39617 "Afghanistan Daylight Time"
  50192 "Alaskan Standard Time"
  50193 "Alaskan Daylight Time"
  55200 "Aleutian Standard Time"
  55201 "Aleutian Daylight Time"
  41968 "Arab Standard Time"
  41969 "Arab Daylight Time"
  42704 "Arabian Standard Time"
  42705 "Arabian Daylight Time"
   8768 "Arabic Standard Time"
   8769 "Arabic Daylight Time"
  19584 "Argentina Standard Time"
  19585 "Argentina Daylight Time"
  34224 "Atlantic Standard Time"
  34225 "Atlantic Daylight Time"
  47024 "AUS Central Standard Time"
  47025 "AUS Central Daylight Time"
  30368 "AUS Eastern Standard Time"
  30369 "AUS Eastern Daylight Time"
  11312 "Azerbaijan Standard Time"
  11313 "Azerbaijan Daylight Time"
  18832 "Azores Standard Time"
  18833 "Azores Daylight Time"
  50896 "Bahia Standard Time"
  50897 "Bahia Daylight Time"
  48720 "Bangladesh Standard Time"
  48721 "Bangladesh Daylight Time"
  56576 "Belarus Standard Time"
  56577 "Belarus Daylight Time"
  25712 "Canada Central Standard Time"
  25713 "Canada Central Daylight Time"
   2000 "Cape Verde Standard Time"
   2001 "Cape Verde Daylight Time"
  29872 "Caucasus Standard Time"
  29873 "Caucasus Daylight Time"
  38928 "Cen. Australia Standard Time"
  38929 "Cen. Australia Daylight Time"
  36656 "Central America Standard Time"
  36657 "Central America Daylight Time"
   8176 "Central Asia Standard Time"
   8177 "Central Asia Daylight Time"
  56336 "Central Brazilian Standard Time"
  56337 "Central Brazilian Daylight Time"
   4896 "Central Europe Standard Time"
   4897 "Central Europe Daylight Time"
   2976 "Central European Standard Time"
   2977 "Central European Daylight Time"
  64416 "Central Pacific Standard Time"
  64417 "Central Pacific Daylight Time"
  17456 "Central Standard Time"
  17457 "Central Daylight Time"
  32800 "Central Standard Time (Mexico)"
  32801 "Central Daylight Time (Mexico)"
    160 "China Standard Time"
    161 "China Daylight Time"
  50384 "Dateline Standard Time"
  50385 "Dateline Daylight Time"
  62720 "E. Africa Standard Time"
  62721 "E. Africa Daylight Time"
  65264 "E. Australia Standard Time"
  65265 "E. Australia Daylight Time"
  64096 "E. South America Standard Time"
  64097 "E. South America Daylight Time"
  58592 "Easter Island Standard Time"
  58593 "Easter Island Daylight Time"
  30800 "Eastern Standard Time"
  30801 "Eastern Daylight Time"
  59424 "Egypt Standard Time"
  59425 "Egypt Daylight Time"
  45952 "Ekaterinburg Standard Time"
  45953 "Ekaterinburg Daylight Time"
  25680 "Fiji Standard Time"
  25681 "Fiji Daylight Time"
  63264 "FLE Standard Time"
  63265 "FLE Daylight Time"
    416 "Georgian Standard Time"
    417 "Georgian Daylight Time"
   7408 "GMT Standard Time"
   7409 "GMT Daylight Time"
  58096 "Greenland Standard Time"
  58097 "Greenland Daylight Time"
  47168 "Greenwich Standard Time"
  47169 "Greenwich Daylight Time"
  24192 "GTB Standard Time"
  24193 "GTB Daylight Time"
   2896 "Haiti Standard Time"
   2897 "Haiti Daylight Time"
  53376 "Hawaiian Standard Time"
  53377 "Hawaiian Daylight Time"
  22400 "India Standard Time"
  22401 "India Daylight Time"
   5568 "Iran Standard Time"
   5569 "Iran Daylight Time"
  47968 "Israel Standard Time"
  47969 "Israel Daylight Time"
  17232 "Jordan Standard Time"
  17233 "Jordan Daylight Time"
  14720 "Kaliningrad Standard Time"
  14721 "Kaliningrad Daylight Time"
  62048 "Korea Standard Time"
  62049 "Korea Daylight Time"
  50768 "Libya Standard Time"
  50769 "Libya Daylight Time"
  16768 "Line Islands Standard Time"
  16769 "Line Islands Daylight Time"
   8832 "Magadan Standard Time"
   8833 "Magadan Daylight Time"
  60896 "Mauritius Standard Time"
  60897 "Mauritius Daylight Time"
  15968 "Middle East Standard Time"
  15969 "Middle East Daylight Time"
  27968 "Montevideo Standard Time"
  27969 "Montevideo Daylight Time"
   2992 "Morocco Standard Time"
   2993 "Morocco Daylight Time"
  34352 "Mountain Standard Time"
  34353 "Mountain Daylight Time"
   7248 "Mountain Standard Time (Mexico)"
   7249 "Mountain Daylight Time (Mexico)"
  21104 "Myanmar Standard Time"
  21105 "Myanmar Daylight Time"
  30704 "N. Central Asia Standard Time"
  30705 "N. Central Asia Daylight Time"
   6208 "Namibia Standard Time"
   6209 "Namibia Daylight Time"
   1216 "Nepal Standard Time"
   1217 "Nepal Daylight Time"
  54896 "New Zealand Standard Time"
  54897 "New Zealand Daylight Time"
   9504 "Newfoundland Standard Time"
   9505 "Newfoundland Daylight Time"
  19056 "North Asia East Standard Time"
  19057 "North Asia East Daylight Time"
    304 "North Asia Standard Time"
    305 "North Asia Daylight Time"
  65008 "Pacific SA Standard Time"
  65009 "Pacific SA Daylight Time"
  11232 "Pacific Standard Time"
  11233 "Pacific Daylight Time"
  15584 "Pacific Standard Time (Mexico)"
  15585 "Pacific Daylight Time (Mexico)"
  28752 "Pakistan Standard Time"
  28753 "Pakistan Daylight Time"
  50224 "Paraguay Standard Time"
  50225 "Paraguay Daylight Time"
  45104 "Romance Standard Time"
  45105 "Romance Daylight Time"
  44560 "Russian Standard Time"
  44561 "Russian Daylight Time"
  40656 "SA Eastern Standard Time"
  40657 "SA Eastern Daylight Time"
  49424 "SA Pacific Standard Time"
  49425 "SA Pacific Daylight Time"
  57120 "SA Western Standard Time"
  57121 "SA Western Daylight Time"
  14592 "Samoa Standard Time"
  14593 "Samoa Daylight Time"
  53728 "SE Asia Standard Time"
  53729 "SE Asia Daylight Time"
  61840 "Singapore Standard Time"
  61841 "Singapore Daylight Time"
  38896 "South Africa Standard Time"
  38897 "South Africa Daylight Time"
  39408 "Sri Lanka Standard Time"
  39409 "Sri Lanka Daylight Time"
  46576 "Syria Standard Time"
  46577 "Syria Daylight Time"
  26432 "Taipei Standard Time"
  26433 "Taipei Daylight Time"
  36016 "Tasmania Standard Time"
  36017 "Tasmania Daylight Time"
  16240 "Tokyo Standard Time"
  16241 "Tokyo Daylight Time"
   7216 "Tonga Standard Time"
   7217 "Tonga Daylight Time"
  24848 "Turkey Standard Time"
  24849 "Turkey Daylight Time"
  47200 "Ulaanbaatar Standard Time"
  47201 "Ulaanbaatar Daylight Time"
  34928 "US Eastern Standard Time"
  34929 "US Eastern Daylight Time"
  37472 "US Mountain Standard Time"
  37473 "US Mountain Daylight Time"
  22000 "Coordinated Universal Time"
  22001 "Coordinated Universal Time"
  51472 "Venezuela Standard Time"
  51473 "Venezuela Daylight Time"
  61376 "Vladivostok Standard Time"
  61377 "Vladivostok Daylight Time"
  65376 "W. Australia Standard Time"
  65377 "W. Australia Daylight Time"
  64976 "W. Central Africa Standard Time"
  64977 "W. Central Africa Daylight Time"
  27696 "W. Europe Standard Time"
  27697 "W. Europe Daylight Time"
  49664 "West Asia Standard Time"
  49665 "West Asia Daylight Time"
  10224 "West Pacific Standard Time"
  10225 "West Pacific Daylight Time"
  40576 "Yakutsk Standard Time"
  40577 "Yakutsk Daylight Time"
}
