/*
 * COM classes for winegstreamer
 *
 * Copyright 2019 Nikolay Sivov
 * Copyright 2019 Zebediah Figura
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#pragma makedep register

[
    helpstring("AVI Splitter"),
    threading(both),
    uuid(1b544c20-fd0b-11ce-8c63-00aa0044b51e)
]
coclass AviSplitter {}

[
    helpstring("MPEG-I Stream Splitter"),
    threading(both),
    uuid(336475d0-942a-11ce-a870-00aa002feab5)
]
coclass MPEG1Splitter {}

[
    helpstring("Wave Parser"),
    threading(both),
    uuid(d51bd5a1-7548-11cf-a520-0080c77ef58a)
]
coclass WAVEParser {}

[
    helpstring("GStreamer splitter"),
    threading(both),
    uuid(f9d8d64e-a144-47dc-8ee0-f53498372c29)
]
coclass Gstreamer_Splitter {}

[
    threading(both),
    uuid(88753b26-5b24-49bd-b2e7-0c445c78c982)
]
coclass VideoProcessorMFT {}
