/*
 * Copyright 1998 Douglas Ridgway
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include <windef.h>

#define IDS_DESCRIPTION                 2
#define IDS_OPEN_META_STRING            3
#define IDI_APPICON                     101
#define IDM_EXIT                        1000
#define IDM_OPEN                        502
#define IDM_SET_EXT_TO_WIN              503
#define IDM_LEFT                        601
#define IDM_RIGHT                       602
#define IDM_UP                          603
#define IDM_DOWN                        604
